import requests
import json
from .utils import all_subclasses, printJSON
import logging

logger = logging.getLogger('prontowork_api')

# base_url = 'http://127.0.0.1:8000/api/'
base_url = 'https://prontowork-alpha-02.appspot.com/api/'

class ProntoWorkAPI:
    def __init__(self, credentials=None, credentials_path=None):
        if credentials:
            self.credentials = credentials
        elif credentials_path:
            with open(credentials_path, 'r') as f:
                self.credentials = json.loads(f.read())
        else:
            raise Exception('Missing credentials, use a credentials dictionary with user and password or the path to '
                            'a json file that contains that information')
        tokens = self.authorize()
        self.access_token = tokens['access']
        self.refresh_token = tokens['refresh']
        # Skills
        self.freelancers = Freelancers(self)
        self.skill_types = SkillTypes(self)
        self.languages = Languages(self)
        self.skills = Skills(self)
        # StreakDB
        self.streak_pipelines = StreakPipelines(self)
        self.streak_fields = StreakFields(self)
        self.streak_stages = StreakStages(self)
        # Task_Input
        self.sources = Sources(self)
        self.campaign_categories = CampaignCategories(self)
        self.campaigns = Campaigns(self)
        self.input_types = InputTypes(self)
        self.inputs = Inputs(self)
        # Task_Output
        self.content_types = ContentTypes(self)
        self.commit_schedules = CommitSchedules(self)
        self.outcomes = Outcomes(self)
        self.output_types = OutputTypes(self)
        self.outputs = Outputs(self)
        self.content_limits = ContentLimits(self)
        # Tasks
        self.task_categories = TaskCategories(self)
        self.tasks = Tasks(self)

    @staticmethod
    def examples():
        print('Examples available https://github.com/prontogit/prontowork_api_examples')

    def authorize(self):
        if 'username' not in self.credentials:
            raise KeyError('Credentials needs a username tag')
        elif 'password' not in self.credentials:
            raise KeyError('Credentials needs a password tag')
        response = requests.post(base_url + "token/", {
            'username': self.credentials['username'], 'password': self.credentials['password']})
        if response.status_code == 401:
            raise NameError('Unauthorized: ' + response.text)
        if response.status_code > 250:
            print(self.credentials)
            raise NameError('Authorization error: ' + response.text)
        return json.loads(response.content)

    def refresh(self):
        response = requests.post(base_url + "refresh/", {'refresh': self.refresh_token})
        if response.status_code == 401:
            raise NameError('Unauthorized: ' + response.text)
        if response.status_code > 250:
            raise NameError('Refresh error: ' + response.text)
        self.access_token = json.loads(response.content)['access']
        return self.access_token

    def api_request(self, method, url, retry_count=0, **kwargs):
        headers = kwargs.pop('headers', {})
        headers['Authorization'] = f'Bearer {self.access_token}'
        headers['Content-Type'] = 'application/json' if 'Content-Type' not in headers else headers['Content-Type']
        headers['Accept'] = 'application/json' if 'Accept' not in headers else headers['Accept']
        response = requests.request(method, url, headers=headers, **kwargs)
        if response.status_code == 401:
            self.refresh()
            if retry_count < 1:
                self.api_request(method, url, retry_count=retry_count + 1, **kwargs)
        elif response.status_code > 250:
            if retry_count < 2:
                self.api_request(method, url, retry_count=retry_count + 1, **kwargs)
            raise Exception('Error ' + str(response.status_code) + ': ' + response.text)
        else:
            try:
                return json.loads(response.text)
            except json.decoder.JSONDecodeError:
                return response.text

    def post(self, url, data):
        return self.api_request('POST', url, data=json.dumps(data))

    def get(self, url, params=None):
        return self.api_request('GET', url, params=params)

    def put(self, url, data):
        return self.api_request('PUT', url, data=json.dumps(data))

    def patch(self, url, data):
        return self.api_request('PATCH', url, data=json.dumps(data))

    def delete(self, url, params=None):
        return self.api_request('DELETE', url, params=params)


class ReadOnlyModel:
    url = base_url

    def __init__(self, pw, url=None):
        self.pw = pw
        if url is not None:
            self.url = url

    def list(self, **kwargs):
        return self.pw.get(url=self.url, params={**kwargs})

    def retrieve(self, pk, **kwargs):
        return self.pw.get(url=self.url + f'{pk}/', params={**kwargs})

    def help(self):
        response = self.pw.get(url=self.url + 'help/')
        printJSON(response)
        return response


class Model(ReadOnlyModel):

    def retrieve_or_create(self, data):
        query = self.list(**data)
        if type(query) is dict:
            query = query['results']
        if not query:
            return self.create(data), True
        elif len(query) == 1:
            return query[0], False
        else:
            raise Exception('Found more than one matching object!')

    def retrieve_one_or_create(self, data):
        query = self.list(**data)
        if type(query) is dict:
            query = query['results']
        if not query:
            return self.create(data), True
        else:
            return query[0], False
        pass

    def create(self, data):
        return self.pw.post(url=self.url, data=data)

    def update(self, pk, data):
        return self.pw.put(url=self.url + f'{pk}/', data=data)

    def destroy(self, pk):
        return self.pw.delete(url=self.url + f'{pk}/')

# Skills
class Freelancers(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'freelancers/'
class SkillTypes(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'skill_types/'
class Languages(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'languages/'
class Skills(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'skills/'

# StreakDB
class StreakPipelines(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'streak_pipelines/'
class StreakFields(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'streak_fields/'
class StreakStages(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'streak_stages/'

# Task_Input
class Sources(Model):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'sources/'
class CampaignCategories(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'campaign_categories/'
class Campaigns(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'campaigns/'
class InputTypes(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'input_types/'
class Inputs(Model):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'inputs/'
        
        

# Task_Output
class ContentTypes(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'content_types/'
class CommitSchedules(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'commit_schedules/'
class Outcomes(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'outcomes/'
class OutputTypes(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'output_types/'

    def generate_outcome_data(self, **kwargs):
        pass
class Outputs(Model):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'outputs/'
class ContentLimits(Model):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'content_limits/'

# Tasks
class TaskCategories(ReadOnlyModel):
    def __init__(self, pw, url=None):
        super().__init__(pw, url)
        self.url = self.url + 'task_categories/'
class Tasks(Model):
    def __init__(self, pw):
        super().__init__(pw)
        self.url = self.url + 'tasks/'

    def task(self, pk):
        return Task(self.pw, pk)

    def retrieve(self, pk):
        task = super().retrieve(pk)
        return self.task(task['pk'])

    def create(self, data):
        task = super().create(data)
        return self.task(task['pk'])

    def update(self, pk, data):
        task = super().update(pk, data)
        return self.task(task['pk'])

    def retrieve_or_create(self, data):
        task, created = super().retrieve_or_create(data)
        return self.task(task['pk']) if not created else task, created

    def retrieve_one_or_create(self, data):
        task, created = super().retrieve_one_or_create(data)
        return self.task(task['pk']) if not created else task, created

    def add_inputs(self, task_pk, input_pks):
        print(self.url + f'{input_pks}/add/')
        if not isinstance(input_pks, list):
            input_pks = [input_pks]
        return self.pw.put(url=self.url + f'{task_pk}/add_inputs/', data={'input_pks': input_pks})

# Nested Tasks
class Task(Tasks):
    def __init__(self, pw, pk):
        super().__init__(pw)
        self.url = self.url + f'{pk}/'
        self.outputs = Outputs(pw, url=self.url)
        self.inputs = Inputs(pw, url=self.url)
        self.pk = pk

    def json(self):
        return self.pw.get(url=self.url)

    def destroy(self):
        return self.pw.delete(url=self.url)

    def add_inputs(self, input_pks):
        if not isinstance(input_pks, list):
            input_pks = [input_pks]
        return self.pw.put(url=self.url + f'add_inputs/', data={'input_pks': input_pks})